"""
ChRIS request module.
"""

import json
import requests
from collection_json import Collection

from.exceptions import ChrisRequestException


class Request(object):
    """
    Http request object.
    """

    def __init__(self, username, password, content_type):
        self.username = username
        self.password = password
        self.content_type = content_type

    def get(self, url, params=None, timeout=30):
        """
        Make a GET request to CUBE.
        """
        headers = {'Content-Type': self.content_type, 'Accept': self.content_type}
        try:
            r = requests.get(url,
                             params=params,
                             auth=(self.username, self.password),
                             timeout=timeout,
                             headers=headers)
        except (requests.exceptions.Timeout, requests.exceptions.RequestException) as e:
            raise ChrisRequestException(str(e))
        return self.get_collection_from_response(r)

    def post(self, url, data, descriptor_file=None, timeout=30):
        """
        Make a POST request to CUBE.
        """
        return self._post_put(requests.post, url, data, descriptor_file, timeout)

    def put(self, url, data, descriptor_file=None, timeout=30):
        """
        Make a PUT request to CUBE.
        """
        return self._post_put(requests.put, url, data, descriptor_file, timeout)

    def delete(self, url, timeout=30):
        """
        Make a DELETE request to CUBE.
        """
        try:
            r = requests.delete(url,
                                auth=(self.username, self.password),
                                timeout=timeout)
        except (requests.exceptions.Timeout, requests.exceptions.RequestException) as e:
            raise ChrisRequestException(str(e))

    def _post_put(self, request_method, url, data, fname=None, timeout=30):
        """
        Internal method to make either a POST or PUT request to CUBE.
        """
        if fname is None:
            headers = {'Content-Type': self.content_type, 'Accept': self.content_type}
            files = None
            data = json.dumps(self.makeTemplate(data))
        else:
            # this is a multipart request
            headers = None
            files = {'fname': fname}
        try:
            r = request_method(url, files=files, data=data,
                               auth=(self.username, self.password),
                               timeout=timeout, headers=headers)
        except (requests.exceptions.Timeout, requests.exceptions.RequestException) as e:
            raise ChrisRequestException(str(e))
        return self.get_collection_from_response(r)

    @staticmethod
    def get_data_from_collection(collection):
        """
        Get the result data dictionary from a collection object.
        """
        result = {'data': [], 'hasNextPage': False, 'hasPreviousPage': False, 'total': 0}
        for item in collection.items:
            item_dict = Request.get_item_descriptors(item)
            result['data'].append(item_dict)
        if Request.get_link_relation_urls(collection, 'next'):
            result['hasNextPage'] = True
        if Request.get_link_relation_urls(collection, 'previous'):
            result['hasPreviousPage'] = True
        if hasattr(collection, 'total'):
            result['total'] = collection.total
        return result

    @staticmethod
    def get_item_descriptors(item):
        """
        Get an item's data (descriptors) in a dictionary.
        """
        item_dict = {}
        # collect the item's descriptors
        for descriptor in item.data:
            item_dict[descriptor.name] = descriptor.value
        return item_dict

    @staticmethod
    def get_link_relation_urls(obj, relation_name):
        """
        Static method to get the list of urls for a link relation in a collection or
        item object.
        """
        return [link.href for link in obj.links if link.rel == relation_name]

    @staticmethod
    def get_collection_from_response(response):
        """
        Static method to get the collection object from a response object.
        """
        content = json.loads(response.text)
        total = content['collection'].pop('total', None)
        collection = Collection.from_json(json.dumps(content))
        if collection.error:
            raise ChrisRequestException(collection.error.message)
        if total is not None:
            collection.total = total
        return collection

    @staticmethod
    def makeTemplate(descriptors_dict):
        """
        Static method to make a Collection+Json template from a regular dictionary whose
        properties are the item descriptors.
        """
        template = {'data': []}
        for key in descriptors_dict:
            template['data'].append({'name': key, 'value': descriptors_dict[key]})
        return {'template': template}



#!/usr/bin/env python3

import  logging
logging.disable(logging.CRITICAL)

import  os
import  sys
import  json
import  socket
import  requests
import  ast
import  pudb

import  pfmisc

# pfstorage local dependencies
from    pfmisc._colors      import  Colors
from    pfmisc.debug        import  debug
from    pfmisc.C_snode      import  *
from    pfstate             import  S

class D(S):
    """
    A derived 'pfstate' class that keeps system state.

    See https://github.com/FNNDSC/pfstate for more information.
    """

    def __init__(self, *args, **kwargs):
        """
        An object to hold some generic/global-ish system state, in C_snode
        trees.
        """
        d_sysArgs   : dict      = {}
        d_CUBE      : dict      = {
                "protocol":     "http",
                "port":         "8000",
                "address":      "%HOSTIP",
                "user":         "chris",
                "password":     "chris1234",
        }
        for k,v in kwargs.items():
            if k == 'args': d_sysArgs   = v
        if 'str_CUBE' in d_sysArgs:
            if len(d_sysArgs['str_CUBE']):
                d_CUBE  = ast.literal_eval("".join(d_sysArgs['str_CUBE'].split()))
        if 'str_CUBEaddress' in d_sysArgs:
            if len(d_sysArgs['str_CUBEaddress']):
                d_CUBE['address']   = d_sysArgs['str_CUBEaddress']
        if 'str_CUBEport' in d_sysArgs:
            if len(d_sysArgs['str_CUBEport']):
                d_CUBE['port']      = d_sysArgs['str_CUBEport']
        self.state_create(
        {
            "CUBE": d_CUBE,
            "self":
            {
                'httpProxy':
                {
                    'use':      False,
                    'httpSpec': ''
                }
            },
            "this":
            {
                'verbosity':    0,
                'colorize':     False
            }
        },
        *args, **kwargs)

class PluginSearch(object):
    """
    A class that interacts with CUBE via the collection+json API
    and is specialized to perform searches on the plugin space.
    """

    def S(self, *args):
        """
        set/get components of the state object
        """
        if len(args) == 1:
            return self.state.T.cat(args[0])
        else:
            self.state.T.touch(args[0], args[1])

    def __init__(self, d_meta, *args, **kwargs):
        """
        Class constructor.
        """
        self.d_args     = vars(*args)
        self.state      = D(
            version     = d_meta['version'],
            name        = d_meta['name'],
            desc        = d_meta['desc'],
            args        = vars(*args)
        )
        if 'verbosity' in self.d_args:  verbosity   = self.d_args['verbosity']
        else:                           verbosity   = 0
        if 'b_syslog' in self.d_args:   b_syslog    = self.d_args['b_syslog']
        else:                           b_syslog    = False
        self.dp         = pfmisc.debug(
            verbosity   = int(verbosity),
            within      = d_meta['name'],
            syslog      = b_syslog,
            colorize    = self.state.T.cat('/this/colorize')
        )
        # Check the IP in the state structure and optionally update
        IP = self.S('/CUBE/address')
        if IP == "%HOSTIP":
            self.S('/CUBE/address', d_meta['defIP'])

    def search_templatize(self):
        """
        Parse the CLI '--using <template>' and return a dictionary
        of parameters to search.

        The <template> value is of form:

            name_exact=<name>,version=<version>,[<key>=<value>,...]

        and is returned as

            {
                'name_exact':   <name>,
                'version':      <version>[,
                <key>:          <value>]
            }

        """
        b_status    : bool      = False
        d_params    : dict      = {}
        str_message : str       = ''
        paramCount  : int       = 0

        if 'str_using' in self.d_args.keys():
            if len(self.d_args['str_using']):
                l_using     = self.d_args['str_using'].split(',')
                for param in l_using:
                    l_keyVal    = param.split('=')
                    d_params[l_keyVal[0]]   = l_keyVal[1]
                    paramCount += 1
                    b_status    = True
                str_message = '%d parameters templatized' % paramCount
            else:
                str_message = "'--using' value is zero length"
        else:
            str_message     = "'--using' not in CLI args"

        return {
            'status':   b_status,
            'message':  str_message,
            'CLI':      self.d_args,
            'params':   d_params
        }

    def search_CUBEAPIcall(self):
        """

        This method implements the actual search logic.

        Caller should check return!

        """
        def dataServiceURL_resolve():
            """
            Based on CLI flags, resolve the actual URL in CUBE.
            """
            str_URL         : str   = ""
            str_id          : str   = ""
            limit           : int   = 50000

            # search across plugins space
            if self.d_args['str_across'] == 'plugins':
                str_URL     = 'api/v1/plugins/search/?limit=%d' % limit
            # search across plugininstance space
            if self.d_args['str_across'] == 'plugininstances':
                str_URL     = 'api/v1/plugins/instances/search/?limit=%d' % \
                                (limit)

            # search across plugin parameter space (given plugin_id)
            if self.d_args['str_across'] == 'parameters':
                if 'plugin_id' in self.d_args['str_using']:
                    str_id  = self.d_args['str_using'].split('=')[1]
                    if str_id.isnumeric():
                        str_URL = 'api/v1/plugins/%s/parameters/?limit=%d' % \
                                    (str_id, limit)

            # search across plugin instance file space (given plugin_inst_id)
            if self.d_args['str_across'] == 'files' or \
               self.d_args['str_across'] == 'links':
                if 'plugin_inst_id' in self.d_args['str_using']:
                    str_id  = self.d_args['str_using'].split('=')[1]
                    if str_id.isnumeric():
                        str_URL = 'api/v1/plugins/instances/%s/files/?limit=%d' % \
                                    (str_id, limit)
            return str_URL

        d_resp              : dict      = {}
        d_templatize        : dict      = self.search_templatize()
        b_status            : bool      = False
        str_message         : str       = 'CUBE API not called because of previous error'
        if d_templatize['status']:
            d_params    = d_templatize['params']
            d_headers           : dict = {
                'Accept':   'application/vnd.collection+json'
            }
            d_resp              : dict  = {}
            str_dataServiceAddr : str   = "%s://%s:%s" % (
                                    self.S('/CUBE/protocol'),
                                    self.S('/CUBE/address'),
                                    self.S('/CUBE/port')
                                )
            str_dataServiceURL  : str   =    dataServiceURL_resolve()
            if len(str_dataServiceURL):
                str_user            : str   = self.S('/CUBE/user')
                str_passwd          : str   = self.S('/CUBE/password')
                str_URL             : str   = '%s/%s' % (
                                        str_dataServiceAddr,
                                        str_dataServiceURL
                                    )
                try:
                    resp = requests.get(
                                        str_URL,
                                        params  = d_params,
                                        auth    = (str_user, str_passwd),
                                        timeout = 30,
                                        headers = d_headers
                            )
                    b_status        = True
                    str_message     = "CUBE call returned a response"
                    d_resp          = resp.json()
                    if 'error' in d_resp['collection']:
                        b_status    = False
                        str_message = d_resp['collection']['error']
                except (requests.exceptions.Timeout,
                        requests.exceptions.RequestException) as e:
                    logging.error(str(e))
                    b_status        = False
                    str_message     = "CUBE call returned some error"
            else:
                str_message         = "Unable to construct a valid service URL. Check if context makes sense."
        return {
            'status':       b_status,
            'templatize':   d_templatize,
            'response':     d_resp,
            'message':      str_message
        }

    def search_desiredReturnFind(self, d_search):
        """
        For a given search response from CUBE, return the
        specific value sought.
        """
        b_status        :   bool    = False
        l_thistarget    :   list    = []
        l_target        :   list    = []
        str_message     :   str     = 'search returned no response'
        b_keysListed    :   bool    = False
        l_keys          :   list    = []
        str_innerList   :   str     = 'data'
        str_innerKey    :   str     = 'name'

        if self.d_args['str_across'] == 'links':
            str_innerList   = 'links'
            str_innerKey    = 'rel'
        if d_search['status']:
            if d_search['response']['collection']['total']:
                for d_hit in d_search['response']['collection']['items']:
                    l_data  :   list = d_hit[str_innerList]
                    if not b_keysListed:
                        for d_el in l_data:
                            l_keys.append(d_el[str_innerKey])
                        b_keysListed = True
                    l_thistarget     = []
                    for str_desired in self.d_args['str_for'].split(','):
                        l_hit   :   list = list(
                                    filter(
                                        lambda info: info[str_innerKey] == str_desired, l_data
                                        )
                                    )
                        if len(l_hit):
                            l_thistarget.append(l_hit[0])
                            b_status    = True
                    if len(self.d_args['str_filterFor']):
                        l_filter = self.d_args['str_filterFor'].split(',')
                        if not len([d for d in l_thistarget     \
                                        for filt in l_filter    \
                                            if d['value'] == filt.strip()]):
                            l_thistarget    = []
                    if len(l_thistarget):
                        l_target.append(l_thistarget)

        if not len(l_target):
            str_message     = "No targets found"
        else:
            str_message     = "%s target(s) found" % len(l_target)

        return {
            'status':   b_status,
            'search':   d_search,
            'target':   l_target,
            'message':  str_message,
            'keys':     l_keys
        }

    def do(self):
        """
        Main entry point to this class.
        """
        d_result    = self.search_desiredReturnFind(self.search_CUBEAPIcall())
        return d_result



"""
ChRIS API client module.
An item in a collection is represented by a dictionary. A collection of items is
represented by a list of dictionaries.
"""

from .request import Request
from .exceptions import ChrisRequestException


class Client(object):
    """
    A ChRIS API client.
    """

    def __init__(self, url, username, password, timeout=30):
        self.url = url
        self.query_url_sufix = 'search/'
        self.username = username
        self.password = password
        self.timeout = timeout
        self.content_type = 'application/vnd.collection+json'

        # urls of the high level API resources
        self.feeds_url = self.url
        self.chris_instance_url = ''
        self.admin_url = ''
        self.files_url = ''
        self.compute_resources_url = ''
        self.plugin_metas_url = ''
        self.plugins_url = ''
        self.plugin_instances_url = ''
        self.pipelines_url = ''
        self.pipeline_instances_url = ''
        self.workflows_url = ''
        self.tags_url = ''
        self.uploaded_files_url = ''
        self.pacs_files_url = ''
        self.service_files_url = ''
        self.file_browser_url = ''
        self.user_url = ''

    def set_urls(self):
        """
        Set the urls of the high level API resources.
        """
        get_url = Request.get_link_relation_urls
        req = Request(self.username, self.password, self.content_type)
        coll = req.get(self.url, None, self.timeout)

        # get urls of the high level API resources
        self.chris_instance_url = self.chris_instance_url or get_url(
            coll, 'chrisinstance')[0]
        self.admin_url = self.admin_url or get_url(coll, 'admin')[0]
        self.files_url = self.files_url or get_url(coll, 'files')[0]
        self.compute_resources_url = self.compute_resources_url or get_url(
            coll, 'compute_resources')[0]
        self.plugin_metas_url = self.plugin_metas_url or get_url(coll, 'plugin_metas')[0]
        self.plugins_url = self.plugins_url or get_url(coll, 'plugins')[0]
        self.plugin_instances_url = self.plugin_instances_url or get_url(
            coll, 'plugin_instances')[0]
        self.pipelines_url = self.pipelines_url or get_url(coll, 'pipelines')[0]
        self.pipeline_instances_url = self.pipeline_instances_url or get_url(
            coll, 'pipeline_instances')[0]
        self.workflows_url = self.workflows_url or get_url(coll, 'workflows')[0]
        self.tags_url = self.tags_url or get_url(coll, 'tags')[0]
        self.uploaded_files_url = self.uploaded_files_url or get_url(
            coll, 'uploadedfiles')[0]
        self.pacs_files_url = self.pacs_files_url or get_url(coll, 'pacsfiles')[0]
        self.service_files_url = self.service_files_url or get_url(
            coll, 'servicefiles')[0]
        self.file_browser_url = self.file_browser_url or get_url(coll, 'filebrowser')[0]
        self.user_url = self.user_url or get_url(coll, 'user')[0]

    def get_feeds(self, search_params=None):
        """
        Get a paginated list of feeds (data descriptors) given query search parameters.
        If no search parameters is given then get the default first page.
        """
        if not self.feeds_url: self.set_urls()
        coll = self._fetch_resource(self.feeds_url, search_params)
        return Request.get_data_from_collection(coll)

    def get_plugins(self, search_params=None):
        """
        Get a paginated list of plugins (data descriptors) given query search
        parameters. If no search parameters is given then get the default first page.
        """
        if not self.plugins_url: self.set_urls()
        coll = self._fetch_resource(self.plugins_url, search_params)
        return Request.get_data_from_collection(coll)

    def get_plugin_by_id(self, id):
        """
        Get a plugin's data (descriptors) given its ChRIS id.
        """
        search_params = {'id': id}
        result = self.get_plugins(search_params)
        if result['data']:
            return result['data'][0]  # plugin ids are unique
        else:
            raise ChrisRequestException(f'Could not find plugin with id {id}')

    def get_plugin_parameters(self, plugin_id, params=None):
        """
        Get a plugin's paginated parameters given its ChRIS id.
        """
        if not self.plugins_url: self.set_urls()
        coll = self._fetch_resource(self.plugins_url, {'id': plugin_id})
        if len(coll.items) == 0:
            raise ChrisRequestException(f'Could not find plugin with id: {plugin_id}.')
        parameters_links = Request.get_link_relation_urls(coll.items[0], 'parameters')
        if parameters_links:
            req = Request(self.username, self.password, self.content_type)
            coll = req.get(parameters_links[0], params, self.timeout) # there can only be a single parameters link
            return Request.get_data_from_collection(coll)
        return {'data': [], 'hasNextPage': False, 'hasPreviousPage': False, 'total': 0}

    def admin_upload_plugin(self, compute_names, fname):
        """
        Upload a plugin representation file and create a new plugin. The fname argument
        can be a string indicating a local file path or a file handler.
        """
        if not self.admin_url: self.set_urls()
        if isinstance(fname, str):
            with open(fname, 'rb') as f:
                file_contents = f.read()
        else:
            file_contents = fname.read()
        req = Request(self.username, self.password, self.content_type)
        data = {'compute_names': compute_names}
        coll = req.post(self.admin_url, data, file_contents, self.timeout)
        result = req.get_data_from_collection(coll)
        return result['data'][0]

    def get_plugin_instances(self, search_params=None):
        """
        Get a paginated list of plugin instances (data descriptors) given query search
        parameters. If no search parameters is given then get the default first page.
        """
        if not self.plugin_instances_url: self.set_urls()
        coll = self._fetch_resource(self.plugin_instances_url, search_params)
        return Request.get_data_from_collection(coll)

    def get_plugin_instance_by_id(self, id):
        """
        Get a plugin instance's data (descriptors) given its ChRIS id.
        """
        search_params = {'id': id}
        result = self.get_plugin_instances(search_params)
        if result['data']:
            return result['data'][0]
        else:
            raise ChrisRequestException(f'Could not find plugin instance with id {id}')

    def create_plugin_instance(self, plugin_id, data):
        """
        Create a plugin instance given the corresponding plugin id and plugin-specific
        data dictionary.
        """
        if not self.plugins_url: self.set_urls()
        coll = self._fetch_resource(self.plugins_url, {'id': plugin_id})
        if len(coll.items) == 0:
            raise ChrisRequestException(f'Could not find plugin with id: {plugin_id}.')
        instances_links = Request.get_link_relation_urls(coll.items[0], 'instances')
        req = Request(self.username, self.password, self.content_type)
        coll = req.post(instances_links[0], data, None, self.timeout)
        result = req.get_data_from_collection(coll)
        return result['data'][0]

    def get_pipelines(self, search_params=None):
        """
        Get a paginated list of pipelines (data descriptors) given query search
        parameters. If no search parameters is given then get the default first page.
        """
        if not self.pipelines_url: self.set_urls()
        coll = self._fetch_resource(self.pipelines_url, search_params)
        return Request.get_data_from_collection(coll)

    def get_pipeline_by_id(self, id):
        """
        Get a pipeline's data (descriptors) given its ChRIS id.
        """
        search_params = {'id': id}
        result = self.get_pipelines(search_params)
        if result['data']:
            return result['data'][0]  # pipeline ids are unique
        else:
            raise ChrisRequestException(f'Could not find pipeline with id {id}')

    def get_pipeline_default_parameters(self, pipeline_id, params=None):
        """
        Get a pipeline's paginated default parameters given its ChRIS id.
        """
        if not self.pipelines_url: self.set_urls()
        coll = self._fetch_resource(self.pipelines_url, {'id': pipeline_id})
        if len(coll.items) == 0:
            raise ChrisRequestException(f'Could not find pipeline with id: '
                                        f'{pipeline_id}.')
        parameters_links = Request.get_link_relation_urls(coll.items[0],
                                                          'default_parameters')
        if parameters_links:
            req = Request(self.username, self.password, self.content_type)
            coll = req.get(parameters_links[0], params, self.timeout)
            return Request.get_data_from_collection(coll)
        return {'data': [], 'hasNextPage': False, 'hasPreviousPage': False, 'total': 0}

    def create_pipeline(self, data):
        """
        Create a pipeline given the data dictionary.
        """
        if not self.pipelines_url: self.set_urls()
        req = Request(self.username, self.password, self.content_type)
        coll = req.post(self.pipelines_url, data, None, self.timeout)
        result = req.get_data_from_collection(coll)
        return result['data'][0]

    def get_workflows(self, search_params=None):
        """
        Get a paginated list of workflows (data descriptors) given query search
        parameters. If no search parameters is given then get the default first page.
        """
        if not self.workflows_url: self.set_urls()
        coll = self._fetch_resource(self.workflows_url, search_params)
        return Request.get_data_from_collection(coll)

    def get_workflow_by_id(self, id):
        """
        Get a workflow's data (descriptors) given its ChRIS id.
        """
        search_params = {'id': id}
        result = self.get_workflows(search_params)
        if result['data']:
            return result['data'][0]
        else:
            raise ChrisRequestException(f'Could not find workflow with id {id}')

    def get_workflow_plugin_instances(self, workflow_id, params=None):
        """
        Get a workflow's paginated list of plugin instances given its ChRIS id.
        """
        if not self.workflows_url: self.set_urls()
        coll = self._fetch_resource(self.workflows_url, {'id': workflow_id})
        if len(coll.items) == 0:
            raise ChrisRequestException(f'Could not find workflow with id: {workflow_id}.')
        parameters_links = Request.get_link_relation_urls(coll.items[0],
                                                          'plugin_instances')
        if parameters_links:
            req = Request(self.username, self.password, self.content_type)
            coll = req.get(parameters_links[0], params, self.timeout)
            return Request.get_data_from_collection(coll)
        return {'data': [], 'hasNextPage': False, 'hasPreviousPage': False, 'total': 0}

    def create_workflow(self, pipeline_id, data):
        """
        Create a workflow given the corresponding pipeline id and pipeline-specific
        data dictionary.
        """
        if not self.pipelines_url: self.set_urls()
        coll = self._fetch_resource(self.pipelines_url, {'id': pipeline_id})
        if len(coll.items) == 0:
            raise ChrisRequestException(f'Could not find pipeline with id: '
                                        f'{pipeline_id}.')
        workflows_links = Request.get_link_relation_urls(coll.items[0], 'workflows')
        req = Request(self.username, self.password, self.content_type)
        coll = req.post(workflows_links[0], data, None, self.timeout)
        result = req.get_data_from_collection(coll)
        return result['data'][0]

    def compute_workflow_nodes_info(self, pipeline_default_parameters,
                                    include_all_defaults=False):
        """
        Helper method to create the nodes_info data structure required to create a
        workflow from a pipeline's default parameters data returned by the
        get_pipeline_default_parameters. If include_all_defaults is set to True
        then non-null parameters are also included in the result.
        """
        pipings_dict = {}
        for default_param in pipeline_default_parameters:
            piping_id = default_param['plugin_piping_id']
            if piping_id not in pipings_dict:
                pipings_dict[piping_id] = {
                    'piping_id': piping_id,
                    'previous_piping_id': default_param['previous_plugin_piping_id'],
                    'compute_resource_name': 'host',
                    'title': default_param['plugin_piping_title'],
                    'plugin_parameter_defaults': []
                }
            if default_param['value'] is None or include_all_defaults:
                pipings_dict[piping_id]['plugin_parameter_defaults'].append(
                    {
                        'name': default_param['param_name'],
                        'default': default_param['value']
                    }
                )
        nodes_info = []
        for piping_id in pipings_dict:
            if not pipings_dict[piping_id]['plugin_parameter_defaults']:
                del pipings_dict[piping_id]['plugin_parameter_defaults']
            nodes_info.append(pipings_dict[piping_id])
        return nodes_info

    def get_pacs_files(self, search_params=None):
        """
        Get a paginated list of PACS files (data descriptors) given query search
        parameters. If no search parameters is given then get the default first page.
        """
        if not self.pacs_files_url: self.set_urls()
        coll = self._fetch_resource(self.pacs_files_url, search_params)
        return Request.get_data_from_collection(coll)

    def register_pacs_file(self, data):
        """
        Register a new PACS file with CUBE.
        """
        if not self.pacs_files_url: self.set_urls()
        req = Request(self.username, self.password, self.content_type)
        coll = req.post(self.pacs_files_url, data, None, self.timeout)
        result = req.get_data_from_collection(coll)
        return result['data'][0]

    def get_service_files(self, search_params=None):
        """
        Get a paginated list of service files (data descriptors) given query search
        parameters. If no search parameters is given then get the default first page.
        """
        if not self.service_files_url: self.set_urls()
        coll = self._fetch_resource(self.service_files_url, search_params)
        return Request.get_data_from_collection(coll)

    def register_service_file(self, data):
        """
        Register a new PACS file with CUBE.
        """
        if not self.service_files_url: self.set_urls()
        req = Request(self.username, self.password, self.content_type)
        coll = req.post(self.service_files_url, data, None, self.timeout)
        result = req.get_data_from_collection(coll)
        return result['data'][0]

    def upload_file(self, upload_path, fname):
        """
        Upload a file to the user's uploads in CUBE. The fname argument can be a string
        indicating a local file path or a file handler.
        """
        if not self.uploaded_files_url: self.set_urls()
        if isinstance(fname, str):
            with open(fname, 'rb') as f:
                file_contents = f.read()
        else:
            file_contents = fname.read()
        req = Request(self.username, self.password, self.content_type)
        data = {'upload_path': upload_path}
        coll = req.post(self.uploaded_files_url, data, file_contents, self.timeout)
        result = req.get_data_from_collection(coll)
        return result['data'][0]

    def get_uploaded_files(self, search_params=None):
        """
        Get a paginated list of uploaded files (data descriptors) given query search
        parameters. If no search parameters is given then get the default first page.
        """
        if not self.uploaded_files_url: self.set_urls()
        coll = self._fetch_resource(self.uploaded_files_url, search_params)
        return Request.get_data_from_collection(coll)

    def delete_uploaded_file(self, id):
        """
        Delete an existing uploaded file.
        """
        if not self.uploaded_files_url: self.set_urls()
        search_params = {'id': id}
        coll = self._fetch_resource(self.uploaded_files_url, search_params)
        file_url = coll.items[0].href
        req = Request(self.username, self.password, self.content_type)
        req.delete(file_url, self.timeout)

    def _fetch_resource(self, url, search_params=None):
        """
        Fetch the collection object of a resource given query search parameters.
        If no search parameters is given then get the default first page.
        """
        req = Request(self.username, self.password, self.content_type)
        if search_params:
            collection = req.get(url + self.query_url_sufix, search_params, self.timeout)
        else:
            collection = req.get(url, None, self.timeout)
        return collection



#!/usr/bin/env python3

import  logging
logging.disable(logging.CRITICAL)

import  os
import  sys
import  json
import  socket
import  requests
import  ast
import  pudb

import  pfmisc
from    chrisclient         import  search
from    argparse            import  Namespace

# pfstorage local dependencies
from    pfmisc._colors      import  Colors
from    pfmisc.debug        import  debug
from    pfmisc.C_snode      import  *
from    pfstate             import  S

class D(S):
    """
    A derived 'pfstate' class that keeps system state.

    See https://github.com/FNNDSC/pfstate for more information.
    """

    def __init__(self, *args, **kwargs):
        """
        An object to hold some generic/global-ish system state, in C_snode
        trees.
        """
        d_sysArgs   : dict      = {}
        d_CUBE      : dict      = {
                "protocol":     "http",
                "port":         "8000",
                "address":      "%HOSTIP",
                "user":         "chris",
                "password":     "chris1234",
        }
        for k,v in kwargs.items():
            if k == 'args': d_sysArgs   = v
        if len(d_sysArgs['str_CUBE']):
            d_CUBE  = ast.literal_eval("".join(d_sysArgs['str_CUBE'].split()))
        if len(d_sysArgs['str_CUBEaddress']):
            d_CUBE['address']   = d_sysArgs['str_CUBEaddress']
        if len(d_sysArgs['str_CUBEport']):
            d_CUBE['port']      = d_sysArgs['str_CUBEport']
        self.state_create(
        {
            "CUBE": d_CUBE,
            "self":
            {
                'httpProxy':
                {
                    'use':      False,
                    'httpSpec': ''
                }
            },
            "this":
            {
                'verbosity':    0,
                'colorize':     False
            }
        },
        *args, **kwargs)

class PluginRun(object):
    """
    A class that interacts with CUBE via the collection+json API
    and is specialized to "run" a plugin via POST requests to the
    backend.
    """

    def S(self, *args):
        """
        set/get components of the state object
        """
        if len(args) == 1:
            return self.state.T.cat(args[0])
        else:
            self.state.T.touch(args[0], args[1])

    def CUBE_IPspec(self):
        """
        Check and update the %HOSTIP if required
        """

        IP = self.S('/CUBE/address')
        if IP == "%HOSTIP":
            self.S('/CUBE/address', self.d_meta['defIP'])

    def __init__(self, d_meta, *args, **kwargs):
        """
        Class constructor.
        """

        # Structures to contain the "CLI" of the plugin
        # to be scheduled/run
        self.d_args         : dict  = vars(*args)
        self.d_CLIargs      : dict  = {}
        self.d_CLIvals      : dict  = {}
        self.d_CLItemplate  : dict  = {}

        # Some additional "massaging" from these CLI
        # to CLI appropriate for the search module:
        self.d_args['str_using']    = self.d_args['str_pluginSpec']
        self.d_args['str_for']      = 'id'
        ns                          = Namespace(**self.d_args)

        # The search module -- used to determine the plugin ID
        # in ChRIS/CUBE
        self.query                  = search.PluginSearch(d_meta, ns)
        self.str_pluginID   : str   = ''

        # Generic "state" data, mostly describing the
        # compute environment in which CUBE exists
        self.d_meta     = d_meta
        self.state      = D(
            version     = d_meta['version'],
            name        = d_meta['name'],
            desc        = d_meta['desc'],
            args        = vars(*args)
        )

        # A debug/print object
        self.dp         = pfmisc.debug(
            verbosity   = int(self.d_args['verbosity']),
            within      = d_meta['name'],
            syslog      = self.d_args['b_syslog'],
            colorize    = self.state.T.cat('/this/colorize')
        )

        # Quick housekeeping
        self.CUBE_IPspec()

    def pluginCLIargs_parse(self):
        """
        Parse the string of CLI args into a dictionary
        structure.
        """
        l_args      :   list    = []
        l_keyval    :   list    = []
        b_add       :   bool    = True
        str_message :   str     = "'--args' is empty!"

        if len(self.d_args['str_args']):
            l_args  = self.d_args['str_args'].split(';')
            for str_keyval in l_args:
                l_keyval    = str_keyval.split('=')
                if len(l_keyval) == 1:
                    key         = l_keyval[0]
                    val         = True
                if len(l_keyval) == 2:
                    (key, val)  = l_keyval
                    val         = val.rstrip('"-=')
                if b_add:
                    # For the set [ " - = ], remove any leading/
                    # trailing hits in the <key> as well as any
                    # trailing hits in the <val>.
                    key         = "".join(key.split())
                    key         = key.strip('"-=')
                    self.d_CLIargs.update({key: val})
        if b_add:
            str_message     = '%d args parsed' % len(l_args)
        return {
            'status':   b_add,
            'message':  str_message,
            'CLIdict':  self.d_CLIargs
        }

    def pluginArgs_CLIvalsFind(self):
        """
        The plugin has a pattern of CLI flags. These flags are associated
        with a variable inside the plugin argparse code. CUBE requires that
        args are POSTed to this variable name, NOT the CLI string.

        This method finds, for each CLI flag, the corresponding value and
        saves this in self.d_CLIvals, key indexed by the name to POST and
        the CLI value.

        Some exceptions to the CLI / name lookup can be defined. These
        are not subject to the indirect value calculation but are passed
        directly to CUBE. Typical exceptions include '--previous_id' along
        with '--title' and '--compute_resource_name'.
        """

        b_status    : bool          = False
        l_hits      : list          = []
        l_directArg : list          = []
        d_val       : dict          = {}
        d_query     : dict          = {}
        d_search    : dict          = {
            'str_for':          'flag,name',
            'str_using':        'plugin_id=%s' % self.str_pluginID,
            'str_across':       'parameters',
            'str_CUBE':         self.d_args['str_CUBE'],
            'str_filterFor':    'key'
        }
        l_directArg = ["previous_id", "title", "compute_resource_name"]
        for key in self.d_CLIargs:
            b_status                        = True
            if key not in l_directArg:
                d_search['str_filterFor']   = '--' + key
                ns                          = Namespace(**d_search)
                self.query                  = search.PluginSearch(self.d_meta, ns)
                d_query                     = self.query.do()
                if len(d_query['target']):
                    l_hits                  = d_query['target'][0]
                    str_userCLIvalue        = self.d_CLIargs[key]
                    d_val = next((el for el in l_hits if el['name'] == 'name'))
                    self.d_CLIvals[d_val['value']] = str_userCLIvalue
            if key in l_directArg:
                self.d_CLIvals[key] = self.d_CLIargs[key]
            # if key == 'previous_id':
            #     self.d_CLIvals['previous_id'] = self.d_CLIargs['previous_id']
        return {
            'status':           b_status,
            'CUBEpluginVals':   self.d_CLIvals
        }

    def pluginArgs_templateCreate(self, d_argsParse):
        """
        Create a collectio+json template from the intenal
        plugin argument dictionary
        """
        b_status            : bool  = True
        d_CUBEpluginVals    : dict  = {}
        self.d_CLItemplate  = {'data' : []}
        str_message         : str   = 'template creation failed due to earlier error'
        keyCount            : int   = 0

        if d_argsParse['status']:
            d_CUBEpluginVals = self.pluginArgs_CLIvalsFind()
            if d_CUBEpluginVals['status']:
                for key in self.d_CLIvals:
                    self.d_CLItemplate['data'].append(
                        {
                            'name':     key,
                            'value':    self.d_CLIvals[key]
                        }
                    )
                    keyCount += 1
        if d_CUBEpluginVals['status']:
            str_message     = '%s key(s) parsed into template' % keyCount
        else:
            str_message     = 'Finding match between CLI flags and CUBE variables failed.'
        return {
            'status':       d_CUBEpluginVals['status'],
            'message':      str_message,
            'template':     self.d_CLItemplate,
            'argsParse':    d_argsParse
        }

    def pluginRun_CUBEAPIcall(self, d_templatize):
        """

        This method POSTs the call to the CUBE API to perform
        the actual plugin execution.

        """
        d_resp              : dict  = {}
        b_status            : bool  = False
        str_message         : str   = 'CUBE API not called because of earlier error'

        if d_templatize['status']:
            d_headers           : dict = {
                'Accept':       'application/vnd.collection+json',
                'Content-Type': 'application/vnd.collection+json'
            }
            str_dataServiceAddr : str  = "%s://%s:%s" % (
                                    self.S('/CUBE/protocol'),
                                    self.S('/CUBE/address'),
                                    self.S('/CUBE/port')
                                )
            str_dataServiceURL  : str  = 'api/v1/plugins/%s/instances/' % \
                                    self.str_pluginID
            str_user            : str  = self.S('/CUBE/user')
            str_passwd          : str  = self.S('/CUBE/password')
            str_URL             : str  = '%s/%s' % (
                                    str_dataServiceAddr,
                                    str_dataServiceURL
                                )
            try:
                resp = requests.post(
                                    str_URL,
                                    data    = json.dumps(
                                        {'template' : self.d_CLItemplate}
                                    ),
                                    auth    = (str_user, str_passwd),
                                    timeout = 30,
                                    headers = d_headers
                        )
                b_status    = True
                str_message     = "CUBE call return a response"
            except (requests.exceptions.Timeout,
                    requests.exceptions.RequestException) as e:
                logging.error(str(e))
                str_message     = "CUBE call return some error"
                raise
            d_resp = resp.json()
        return {
            'status':       b_status,
            'templatize':   d_templatize,
            'response':     d_resp,
            'message':      str_message
        }

    def pluginInstanceID_find(self, d_run):
        """
        Once a job has been POSTed, parse the return from
        CUBE for the plugin instance ID.
        """
        b_status        :   bool    = False
        l_thistarget    :   list    = []
        l_target        :   list    = []
        str_message     :   str     = 'No instance ID found due to earlier error'

        if d_run['status']:
            if 'error' in d_run['response']['collection'].keys():
                str_message = d_run['response']['collection']['error']['message']
            if 'items' in d_run['response']['collection'].keys():
                for d_hit in d_run['response']['collection']['items']:
                    l_data  :   list = d_hit['data']
                    l_thistarget     = []
                    for str_desired in self.d_args['str_for'].split(','):
                        l_hit   :   list = list(
                                    filter(
                                        lambda info: info['name'] == str_desired, l_data
                                        )
                                    )
                        if len(l_hit):
                            l_thistarget.append(l_hit[0])
                            b_status    = True
                    l_target.append(l_thistarget)
                str_message     = '%d instance ID(s) found' % len(l_target)
        return {
            'status':   b_status,
            'run':      d_run,
            'target':   l_target,
            'message':  str_message
        }

    def do(self):
        """
        Main entry point to this class.
        """
        b_status    :   bool    = False
        str_message :   str     = ''
        d_query     :   dict    = {}
        d_targetID  :   dict    = {}
        d_run       :   dict    = {}

        # First, find the plugin ID for the desired plugin
        # to run
        d_query             = self.query.do()
        if len(d_query['target']) >= 1:
            if len(d_query['target']) > 1:
                str_message     += "multiple hits found, using first\n"
            d_targetID          = d_query['target'][0][0]
            self.str_pluginID   = d_targetID['value']

            # and now, run it!
            d_run:  dict    = self.pluginInstanceID_find(
                                self.pluginRun_CUBEAPIcall(
                                    self.pluginArgs_templateCreate(
                                        self.pluginCLIargs_parse()
                                    )
                                )
                            )
            if d_run['status']:
                str_message += 'plugin scheduled successfully'
                b_status    = True
            else:
                str_message += 'plugin run NOT scheduled -- some error returned'
        else:
            str_message     = "no valid plugin found"
        return {
            'status':       b_status,
            'query':        d_query,
            'run':          d_run,
            'message':      str_message
        }






from os import path
from setuptools import setup

with open(path.join(path.dirname(path.abspath(__file__)), 'README.rst')) as f:
    readme = f.read()

setup(
      name             =   'python-chrisclient',
      version          =   '2.7.0',
      description      =   '(Python) client for the ChRIS API',
      long_description =   readme,
      author           =   'FNNDSC',
      author_email     =   'dev@babymri.org',
      url              =   'https://github.com/FNNDSC/python-chrisclient',
      packages         =   ['chrisclient'],
      install_requires =   ['requests>=2.21.0', 'collection-json>=0.1.1', 'pudb', 'pfstate', 'pfmisc', 'webob'],
      test_suite       =   'nose.collector',
      tests_require    =   ['nose'],
      scripts          =   ['bin/chrisclient', 'bin/chrispl-run', 'bin/chrispl-search'],
      license          =   'MIT',
      zip_safe         =   False,
      python_requires  =   '>=3.7'
)
